# pfeiffer_vacuum_protocol - Python interface to Pfeiffer vacuum gauges
# Copyright (C) 2020 Christopher M. Pierce (contact@chris-pierce.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from enum import Enum
import io

# Error states for vacuum gauges
class ErrorCode(Enum):
    NO_ERROR = 1
    DEFECTIVE_TRANSMITTER = 2
    DEFECTIVE_MEMORY = 3

def _send_data_request(s, addr, param_num):
    c =  "{:03d}00{:03d}02=?".format(addr, param_num)
    c += "{:03d}\r".format(sum([ord(x) for x in c])%256)
    s.write(c.encode())

def _send_control_command(s, addr, param_num, data_str):
    c =  "{:03d}10{:03d}{:02d}{:s}".format(addr, param_num, len(data_str), data_str)
    c += "{:03d}\r".format(sum([ord(x) for x in c])%256)
    return s.write(c.encode())

def _read_gauge_response(s):
    # Read until newline or we stop getting a response
    r = ""
    for _ in range(64):
        c = s.read(1)
        if(c == b""):
            break
        r += c.decode("ascii")
        if(c == b"\r"):
            break

    # Check the length
    if(len(r) < 14):
        raise ValueError("gauge response too short to be valid")

    # Check it is terminated correctly
    if(r[-1] != "\r"):
        raise ValueError("gauge response incorrectly terminated")

    # Evaluate the checksum
    if(int(r[-4:-1]) != (sum([ord(x) for x in r[:-4]])%256)):
        raise ValueError("invalid checksum in gauge response")

    # Pull out the address
    addr      = int(r[:3])
    rw        = int(r[3:4])
    param_num = int(r[5:8])
    data      =  r[10:-4]

    # Check for errors
    if(data == "NO_DEF"):
        raise ValueError("undefined parameter number")
    if(data == "_RANGE"):
        raise ValueError("data is out of range")
    if(data == "_LOGIC"):
        raise ValueError("logic access violation")

    # Return it
    return addr, rw, param_num, data

def read_error_code(s, addr):
    _send_data_request(s, addr, 303)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    if(raddr!=addr or rw!=1 or rparam_num!=303):
        raise ValueError("invalid response from gauge")

    if(rdata == "000000"):
        return ErrorCode.NO_ERROR
    elif(rdata == "Err001"):
        return ErrorCode.DEFECTIVE_TRANSMITTER
    elif(rdata == "Err002"):
        return ErrorCode.DEFECTIVE_MEMORY
    else:
        raise ValueError("unexpected error code from gauge")

def read_software_version(s, addr):
    _send_data_request(s, addr, 312)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    if(raddr!=addr or rw!=1 or rparam_num!=312):
        raise ValueError("invalid response from gauge")

    return (int(rdata[0:2]), int(rdata[2:4]), int(rdata[4:]))

def read_gauge_type(s, addr):
    _send_data_request(s, addr, 349)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    if(raddr!=addr or rw!=1 or rparam_num!=349):
        raise ValueError("invalid response from gauge")

    if(rdata == "    A1"):
        return "CPT 100"
    elif(rdata == "    A2"):
        return "RPT 100"
    elif(rdata == "    A3"):
        return "PPT 100"
    elif(rdata == "    A4"):
        return "HPT 100"
    elif(rdata == "    A5"):
        return "MPT 100"
    else:
        raise ValueError("unrecognized gauge type")

def read_pressure(s, addr):
    _send_data_request(s, addr, 740)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    if(raddr!=addr or rw!=1 or rparam_num!=740):
        raise ValueError("invalid response from gauge")

    # Convert to a float
    mantissa = int(rdata[:4])
    exponent = int(rdata[4:])
    return float(mantissa*10**(exponent-26))

def write_pressure_setpoint(s, addr, val):
    # Format the data
    data =  "{:03d}".format(val)
    _send_control_command(s, addr, 741, data)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    # Check the response
    if(raddr!=addr or rw!=1 or rparam_num!=741):
        raise ValueError("invalid response from gauge")

    if(rdata != data):
        raise ValueError("invalid acknowledgment from gauge")

def read_correction_value(s, addr):
    _send_data_request(s, addr, 742)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    if(raddr!=addr or rw!=1 or rparam_num!=742):
        raise ValueError("invalid response from gauge")

    return float(rdata)/100

def write_correction_value(s, addr, val):
    # Format the data
    data =  "{:06d}".format(int(val*100))
    _send_control_command(s, addr, 742, data)
    raddr, rw, rparam_num, rdata = _read_gauge_response(s)

    # Check the response
    if(raddr!=addr or rw!=1 or rparam_num!=742):
        raise ValueError("invalid response from gauge")

    if(rdata != data):
        raise ValueError("invalid acknowledgment from gauge")



# pfeiffer_vacuum_protocol - Python interface to Pfeiffer vacuum gauges
# Copyright (C) 2020 Christopher M. Pierce (contact@chris-pierce.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
import pfeiffer_vacuum_protocol.mock as mock
import pfeiffer_vacuum_protocol as pvp

class TestPVP(unittest.TestCase):
    def test_read_error_code(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = pvp.read_error_code(s, 1)
        self.assertEqual(r, pvp.ErrorCode.NO_ERROR)
        s = mock.Serial(mock.PPT100(err_state=pvp.ErrorCode.DEFECTIVE_TRANSMITTER), "COM1")
        r = pvp.read_error_code(s, 1)
        self.assertEqual(r, pvp.ErrorCode.DEFECTIVE_TRANSMITTER)
        s = mock.Serial(mock.PPT100(err_state=pvp.ErrorCode.DEFECTIVE_MEMORY), "COM1")
        r = pvp.read_error_code(s, 1)
        self.assertEqual(r, pvp.ErrorCode.DEFECTIVE_MEMORY)

    def test_read_software_version(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = pvp.read_software_version(s, 1)
        self.assertEqual(r, (1,1,0))

    def test_read_gauge_type(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = pvp.read_gauge_type(s, 1)
        self.assertEqual(r, "PPT 100")

    def test_read_pressure(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = pvp.read_pressure(s, 1)
        self.assertEqual(r, 1.0)

    def test_write_pressure_setpoint(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        pvp.write_pressure_setpoint(s, 1, 0)
        pvp.write_pressure_setpoint(s, 1, 1)
        with self.assertRaises(ValueError) as context:
            pvp.write_pressure_setpoint(s, 1, 2)

    def test_read_correction_value(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = pvp.read_correction_value(s, 1)
        self.assertEqual(r, 1.0)

    def test_write_correction_value(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        pvp.write_correction_value(s, 1, 1.0)

if __name__ == '__main__':
    unittest.main()



# pfeiffer_vacuum_protocol - Python interface to Pfeiffer vacuum gauges
# Copyright (C) 2020 Christopher M. Pierce (contact@chris-pierce.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import io
from .pfeiffer_vacuum_protocol import ErrorCode

# Pulled from pySerial
PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 'N', 'E', 'O', 'M', 'S'
STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO = (1, 1.5, 2)
FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS = (5, 6, 7, 8)

def to_bytes(seq):
    """convert a sequence to a bytes type"""
    if isinstance(seq, bytes):
        return seq
    elif isinstance(seq, bytearray):
        return bytes(seq)
    elif isinstance(seq, memoryview):
        return seq.tobytes()
    elif isinstance(seq, str):
        raise TypeError('unicode strings are not supported, please encode to bytes: {!r}'.format(seq))
    else:
        # handle list of integers and bytes (one or more items) for Python 2 and 3
        return bytes(bytearray(seq))

class Serial(io.RawIOBase):
    """\
    Mockup of a serial port named COM1 connected to a pfeiffer device
    """
    # Default values copied from pySerial
    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200, 230400, 460800, 500000,
                 576000, 921600, 1000000, 1152000, 1500000, 2000000, 2500000,
                 3000000, 3500000, 4000000)
    BYTESIZES = (FIVEBITS, SIXBITS, SEVENBITS, EIGHTBITS)
    PARITIES = (PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE)
    STOPBITS = (STOPBITS_ONE, STOPBITS_ONE_POINT_FIVE, STOPBITS_TWO)

    def __init__(self,
                 connected_device,
                 port=None,
                 baudrate=9600,
                 bytesize=EIGHTBITS,
                 parity=PARITY_NONE,
                 stopbits=STOPBITS_ONE,
                 timeout=None,
                 xonxoff=False,
                 rtscts=False,
                 write_timeout=None,
                 dsrdtr=False,
                 inter_byte_timeout=None,
                 exclusive=None,
                 **kwargs):
        """\
        Initializes the com port object
        """
        self.buffer = b""
        self.dev = connected_device
        self.baudrate = baudrate

    def flush(self):
        self.buffer = b""

    def write(self, output):
        self.buffer += self.dev.get_response(to_bytes(output))
        return len(output)

    def read(self, readlen):
        if(self.baudrate != 9600):
            return b""

        ret = self.buffer[:readlen]
        self.buffer = self.buffer[readlen:]
        return ret

    def readinto(self, b):
        data = self.read(len(b))
        n = len(data)
        b[:n] = data
        return n

    def readable(self):
        return True

    def writable(self):
        return True

    def seekable(self):
        return False

class PPT100:
    """\
    Mockup of the Pfeiffer vacuum gauge model PPT 100
    """
    def __init__(self, address=1, err_state=ErrorCode.NO_ERROR):
        self.address = address
        self.err_state = err_state

    def get_response(self, bin_str):
        # Convert to a unicode str
        in_str = bin_str.decode("ascii")

        # If it doesn't end in a carriage return, exit
        if(in_str[-1] != '\r'):
            return b""

        # Validate the checksum and exit if bad
        if(int(in_str[-4:-1]) != (sum([ord(x) for x in in_str[:-4]])%256)):
            return b""

        # Get the address, and exit if it isn't correct
        if(int(in_str[:3]) != 1):
            return b""

        # Get the data length and return if it's wrong
        if(len(in_str)-14 != int(in_str[8:10])):
            return b""

        # Get the operation type (read/write)
        op_type = int(in_str[3])

        # Get the paramter number
        param_num = int(in_str[5:8])

        # If we are reading
        if(op_type == 0):
            # Check that the data length is right and exit if not
            if(int(in_str[8:10]) != 2):
                return b""

            # Check that the data is =? exit if it isn't
            if(in_str[10:12] != "=?"):
                return b""

            # If it is error code, return the right one
            if(param_num == 303):
                if(self.err_state == ErrorCode.NO_ERROR):
                    return b'0011030306000000014\r'

                elif(self.err_state == ErrorCode.DEFECTIVE_TRANSMITTER):
                    return b'0011030306Err001168\r'

                elif(self.err_state == ErrorCode.DEFECTIVE_MEMORY):
                    return b'0011030306Err002169\r'

                else:
                    raise ValueError("unknown error state")

            # Or, if it's version number, return it
            elif(param_num == 312):
                return b'0011031206010100016\r'

            # Or, if it's the component name, return it
            elif(param_num == 349):
                return b'0011034906    A3236\r'

            # Or, if it's pressure, return it
            elif(param_num == 740):
                return b'0011074006100023025\r'

            # Or, if it's the correction value, return it
            elif(param_num == 742):
                return b'0011074206000100022\r'

            # If it wasn't any of these, return the error code
            else:
                resp = "00110{:03d}06NO_DEF".format(param_num)
                resp += "{:03d}\r".format(sum([ord(x) for x in resp])%256)
                return resp.encode()

        # Or, if it's write
        else:
            # If it was set point
            if(param_num == 741):
                # Check the datatype
                if(int(in_str[8:10]) != 3):
                    return b'0011074106NO_DEF191\r'

                # Check bounds and return error if we're out
                val = int(in_str[10:13])
                if(val<0 or val>1):
                    return b'0011074106_RANGE192\r'

                # Return the confirmation
                return in_str.encode()

            # Or, if it was the correction value
            elif(param_num == 742):
                # Check the datatype
                if(int(in_str[8:10]) != 6):
                    return b'0011074206NO_DEF192\r'

                # Return the confirmation
                return in_str.encode()

            # If it wasn't a valid parameter, return an error
            else:
                return b'0011074206NO_DEF192\r'



# pfeiffer_vacuum_protocol - Python interface to Pfeiffer vacuum gauges
# Copyright (C) 2020 Christopher M. Pierce (contact@chris-pierce.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
import pfeiffer_vacuum_protocol.mock as mock
from pfeiffer_vacuum_protocol import ErrorCode
import io

class TestSerial(unittest.TestCase):
    def test_write(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        r = s.write(b"test")
        self.assertEqual(r, 4)

    def test_wrong_type(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        with self.assertRaises(TypeError) as context:
            s.write("blah blah blah")

    def test_read(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        s.write(b"0010073902=?114\r")
        r = s.read(1)
        self.assertEqual(r, b"0")

    def test_io_readline(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        sio = io.TextIOWrapper(io.BufferedRWPair(s, s), newline="\r")
        sio.write("0010073902=?114\r")
        sio.flush()
        r = sio.readline()
        self.assertEqual(r, "0011073906NO_DEF198\r")

    def test_wrong_baudrate(self):
        s = mock.Serial(mock.PPT100(), "COM1", 19200)
        s.write(b"0010073902=?114\r")
        r = s.read(1)
        self.assertEqual(r, b"")

    def test_flush(self):
        s = mock.Serial(mock.PPT100(), "COM1")
        s.flush()

class TestPPT100(unittest.TestCase):
    def test_no_cr(self):
        g = mock.PPT100()
        r = g.get_response(b"0010074002=?106")
        self.assertEqual(r, b"")

    def test_no_def(self):
        g = mock.PPT100()
        r = g.get_response(b"0010073902=?114\r")
        self.assertEqual(r, b'0011073906NO_DEF198\r')

    def test_no_device(self):
        g = mock.PPT100()
        r = g.get_response(b"0020074002=?107\r")
        self.assertEqual(r, b"")

    def test_bad_cs(self):
        g = mock.PPT100()
        r = g.get_response(b"0010074002=?107\r")
        self.assertEqual(r, b"")

    def test_actual_err_none(self):
        g = mock.PPT100()
        r = g.get_response(b"0010030302=?101\r")
        self.assertEqual(r, b'0011030306000000014\r')

    def test_actual_err_trans(self):
        g = mock.PPT100(err_state=ErrorCode.DEFECTIVE_TRANSMITTER)
        r = g.get_response(b"0010030302=?101\r")
        self.assertEqual(r, b'0011030306Err001168\r')

    def test_actual_err_mem(self):
        g = mock.PPT100(err_state=ErrorCode.DEFECTIVE_MEMORY)
        r = g.get_response(b"0010030302=?101\r")
        self.assertEqual(r, b'0011030306Err002169\r')

    def test_software_vers(self):
        g = mock.PPT100()
        r = g.get_response(b"0010031202=?101\r")
        self.assertEqual(r, b'0011031206010100016\r' )

    def test_component_name(self):
        g = mock.PPT100()
        r = g.get_response(b"0010034902=?111\r")
        self.assertEqual(r, b'0011034906    A3236\r')

    def test_pressure_in(self):
        g = mock.PPT100()
        r = g.get_response(b"0010074002=?106\r")
        self.assertEqual(r, b'0011074006100023025\r' )

    def test_pressure_setpoint(self):
        g = mock.PPT100()
        r = g.get_response(b"0011074103001130\r")
        self.assertEqual(r, b'0011074103001130\r')
        r = g.get_response(b"0011074103000129\r")
        self.assertEqual(r, b'0011074103000129\r')

    def test_pressure_setpoint_range(self):
        g = mock.PPT100()
        r = g.get_response(b"0011074103002131\r")
        self.assertEqual(r, b'0011074106_RANGE192\r')
        r = g.get_response(b"0011074103003132\r")
        self.assertEqual(r, b'0011074106_RANGE192\r')
        r = g.get_response(b"0011074103011131\r")
        self.assertEqual(r, b'0011074106_RANGE192\r')
        r = g.get_response(b"0011074103101131\r")
        self.assertEqual(r, b'0011074106_RANGE192\r')

    def test_pirani_correction_write(self):
        g = mock.PPT100()
        r = g.get_response(b"0011074206000100022\r")
        self.assertEqual(r, b'0011074206000100022\r')

    def test_pirani_correction_read(self):
        g = mock.PPT100()
        r = g.get_response(b"0010074202=?108\r")
        self.assertEqual(r, b'0011074206000100022\r' )

    def test_wrong_datalen_read(self):
        g = mock.PPT100()
        r = g.get_response(b"0010074201=044\r")
        self.assertEqual(r, b"")

    def test_wrong_datalen_write(self):
        g = mock.PPT100()
        r = g.get_response(b"0011074203100131\r")
        self.assertEqual(r, b'0011074206NO_DEF192\r' )

    def test_wrong_data_read(self):
        g = mock.PPT100()
        r = g.get_response(b"00100742021?096\r")
        self.assertEqual(r, b"")
        r = g.get_response(b"0010074201=?107\r")
        self.assertEqual(r, b"")

    def test_wrong_data_write(self):
        g = mock.PPT100()
        r = g.get_response(b"0011074206000133\r")
        self.assertEqual(r, b"")

if __name__ == '__main__':
    unittest.main()



# pfeiffer_vacuum_protocol - Python interface to Pfeiffer vacuum gauges
# Copyright (C) 2020 Christopher M. Pierce (contact@chris-pierce.com)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import setuptools

setuptools.setup(
        name='pfeiffer-vacuum-protocol',
        version='0.2',
        description=' Python interface to Pfeiffer vacuum gauges',
        author='Christopher M. Pierce',
        author_email='contact@chris-pierce.com',
        python_requires='>=3.5',
        packages=setuptools.find_packages(),
        long_description_content_type="text/markdown",
        install_requires = [
            'pyserial',
        ],
        license = "GNU Lesser General Public License v3 (LGPLv3)",
        classifiers = [
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Development Status :: 4 - Beta",
            "Operating System :: OS Independent",
        ],
    )
